#!BPY

""" Registration info for Blender menus:
Name: 'Ldraw'
Blender: 233
Group: 'Import'
Tip: 'Import Ldraw (LEGO Cad) files.'
"""

#////////////////////////////////////// LDRAW ///////////////////////////////////*
#/////
#/////	SCRIPT: LDRAW_IMPORT.PY
#/////	VERSION: 0.03.00
#/////
#/////	AUTHORS:	Alessandro V. 2004 - alv67@katamail.com
#/////                  Alex Lindeijer
#/////
#/////	REQUIRMENTS:	Blender 2.33+ (don't know if it works in version 2.32)
#/////
#/////  REFERENCES:     LDraw.org Official Specifications
#/////                  - LDraw File Format          (May 29, 2003)
#/////                  - MPD Language Extension     (May 29, 2003)
#/////                  - LDraw Language Extension for Clipping
#/////                                               (Sep 22, 2003)
#////                   - LDraw Color Chart          (Sep 05, 2003)
#/////
#/////	TODO:	        - Adjust crossed vertex faces
#/////                  - Read Multi-Part DAT files (MPD extension)
#/////                      "0 FILE model.ldr"  Multi-part DAT or MPD
#/////                  - Import       "Line"  ("2") command (if needed)
#/////                      "Conditiona-line"  ("5") command (if needed)
#/////                  - Improve color import: Correct color, Multicolor Parts
#/////                  - Usage of Blender 2.34 and improve code
#/////
#/////	LICENSE:        The script is under GPL and feel free to use it.
#/////
#/////
#/////- REVISION HISTORY --------------------------------------------------------
#/////
#/////
#/////  Version 0.03.00 - 18 august 2004
#/////  --------------------------------
#/////  by Alessandro V. - alv67@katamail.com
#/////    Added : - Transperent colours
#/////    Fixes : - Corrected 'black' faces by use of 'TwoFaces' mesh
#/////    Code  : - Totaly rewritted code, introduced two classes LegoFile
#/////              and LegoObject that contains all the necessary code
#/////            - Eliminated all unneeded string functions
#/////              using string object methods instead:
#/////               string.split()    string.upper()
#/////            - Eliminated all unneeded matrix functions
#/////              using matrix object methods and functions
#/////               matrix1 * matrix2   VecMultMat1(v,m)
#/////
#/////  Version 0.02.01 - 09 august 2004
#/////  --------------------------------
#/////  by Alex Lindeijer
#/////    Fixes : - added legocolor to file
#/////
#/////  Version 0.02.00 - 08 august 2004
#/////  --------------------------------
#/////  by Alex Lindeijer
#/////    Added : - every part and model (in PARTS and MODELS directory is an object)
#/////            - mesh for same partnumber is reused.
#/////    Fixes : - Correct coords: blender (x,y,z)= ldraw(x,z.-y)
#/////            - no double vertexes
#/////
#/////  Version 0.01.01 - 04 july 2004
#/////  ------------------------------
#/////  by Alessandro V. - alv67@katamail.com
#/////    Fixes : - Corrected bad rotation and translation of some objects due
#/////              to a not good code in transposing coordinate system.
#/////              Now object are imported whith original (Ldraw) axes.
#/////     Code : - Introduced the Split() function and restructured software
#/////              using this function and deleting CVertex, CValue classes
#/////
#/////  Version 0.01.00 - 29 june 2004
#/////  ------------------------------
#/////    Initial release by Alessandro V. - alv67@katamail.com
#/////
#////////////////////////////////////// LDRAW ///////////////////////////////////*


import Blender
from Blender import NMesh, Material, Texture, Scene, Lamp, Image, sys
from Blender.Draw import *
from Blender.BGL import *
from Blender.Mathutils import *
LDrawPath="C:\Ldraw"      # LDRAW base installation directory
Scale=0.05                # 1 Lego unit in Blender Unit


global mtxIdentity        # Null effect translation matrix

mtxIdentity = Blender.Mathutils.Matrix()
mtxIdentity.resize4x4()
mtxIdentity.identity()

EVENT_PATHCHANGE=		1
EVENT_IMPORT=			2
EVENT_IMPORT_MATS=		3
EVENT_QUIT=			4

#= [CLASS] LegoFile ======================================================
#
class LegoFile:
    def __init__(self, filename, searchPath = "") :
        if 'S\\' in filename:
            self.Name = filename
        else:
            self.Name = sys.basename(filename)
        self.isPart = False
        self.FileName = None
        self.Path = ""
        try:
            aFile = open(filename, "rb")
        except IOError:
            # try in the search path
            filename = searchPath + sys.sep + self.Name
            try:
                aFile = open(filename, "rb")
            except IOError:
                # try in the MODEL path
                filename = LDrawPath + sys.sep + \
                           "MODELS" + sys.sep + self.Name
                try:
                    aFile = open(filename, "rb")
                except IOError:
                    self.isPart = True
                    # Search for part in Ldraw PARTS directory
                    filename = LDrawPath + sys.sep + \
                               "PARTS" + sys.sep + self.Name
                    try:
                        aFile = open(filename, "rb")
                    except IOError:
                        # Search for part in Ldraw P directory
                        filename = LDrawPath + sys.sep + \
                                   "P" + sys.sep + self.Name
                        try:
                            aFile = open(filename, "rb")
                        except IOError:
                            print("Unable to find part file %s!" \
                                  % (self.Name))
                            return None
        self.FileName = filename
        self.Path = sys.dirname(filename)
        aFile.close()
    def __str__(self):
         retstr =  " CLASS : LegoFile \n"
         retstr += "   FileName : %s \n" % (self.FileName)
         retstr += "   Name     : %s \n" % (self.Name)
         retstr += "   Path     : %s \n" % (self.Path)
         retstr += "   isPart   : %s \n" % (self.isPart)
         return retstr
#=========================================================================

#=[CLASS] LegoObject =====================================================
#
class LegoObject:
    def __init__(self, type, name=""):
        if type != 'Model' and type != 'Part':
            print "LegoObject accept only 'Model' and 'Part' types"
            return null
        self.Type = type
        self.Name = name
        self.Mesh = Blender.NMesh.New(self.Name)
        self.Mesh.setMaxSmoothAngle(45)
        self.Mesh.setMode("AutoSmooth", "TwoSided")
        if self.Type == 'Model':
            self.Object = Blender.Object.New('Mesh',self.Name)
            self.Object.link(self.Mesh)
        self.lstVertex = []
        self.MeshMaterials = {}

    def ImportFile(self, lFile, matrix=mtxIdentity, pcolor=16, Invert=False):
        InvertNext = False
        CW = Invert
        # Open the DAT file
        try:
            aFile = open(lFile.FileName, "rb")
        except IOError:
            print("Unable to find part file %s!" % (lFile.FileName))
            return
        # Once the file is opened, we read all the lines
        for line in aFile:
            cmdLine = line.split()
            if cmdLine == []:
               pass
            elif cmdLine[0]=='0':  # Remark
               if "BFC" in cmdLine:
                  if " CCW" in line:
                     # Do nothing as sometimes setting CW or CCW is not completely OK
                     # Patched setting 'TwoFaces' mesh
                     pass
                  if " CW" in line:
                     # Do nothing as sometimes setting CW or CCW is not completely OK
                     # Patched setting 'TwoFaces' mesh
                     pass
                  if "INVERTNEXT" in cmdLine:
                      InvertNext = True
               pass
            elif cmdLine[0]=='1':  # Part File
               self.AddPart(cmdLine, matrix, pcolor, InvertNext)
               InvertNext = False
               pass
            elif cmdLine[0]=='2':  # Line
               pass
            elif cmdLine[0]=='3':  # Triangle
               if self.Type == 'Part':
                   self.AddTriangle(cmdLine, matrix,pcolor, CW)
               else:
                   self.AddTriangle(cmdLine, matrix,16, CW)
            elif cmdLine[0]=='4':  # Quadrilateral
               if self.Type == 'Part':
                 self.AddQuadrilateral(cmdLine, matrix, pcolor, CW)
               else:
                 self.AddQuadrilateral(cmdLine, matrix, 16, CW)
            elif cmdLine[0]=='5':  # Optional-Line
               InvertNext = False
               pass
            else :
               InvertNext = False
               pass
        aFile.close()
        self.Mesh.update()

    def AddPart(self, cmd, matrix=mtxIdentity, color=16, Invert=False ):
        if cmd[0] != '1':       # Check if line contains correct command
            return
        if len(cmd) != 15:      # Check for correct number of parameters
            return
        # Get all the command parameters
        newcolor = int(cmd[1])
        if newcolor == 16:
            newcolor = color
        mtxTranf = Matrix([float(cmd[ 5]), float(cmd[ 8]), float(cmd[11]),0],
                          [float(cmd[ 6]), float(cmd[ 9]), float(cmd[12]),0],
                          [float(cmd[ 7]), float(cmd[10]), float(cmd[13]),0],
                          [float(cmd[ 2]), float(cmd[ 3]), float(cmd[ 4]),1])
        mtxTranf *= matrix
        partName  = cmd[14].upper()
        LFile = LegoFile(partName)

        # Check if needed to create a new lego object or integrate
        # new part in the current object mesh
        if self.Type == 'Part':
            self.ImportFile(LFile, mtxTranf, newcolor, Invert)
        elif self.Type == 'Model':
            scene = Blender.Scene.getCurrent()
            if LFile.isPart:
                newObj = Blender.Object.New('Mesh',partName)
                oldMesh = Blender.NMesh.GetRaw(LFile.Name)
                if oldMesh != None:
                # This part is present into Blender
                    newMesh=oldMesh
                else:
                    newLObj = LegoObject('Part', LFile.Name)
                    newLObj.ImportFile(LFile, Invert=Invert)
                    newMesh = newLObj.Mesh
                    newMesh.addMaterial(Material.Get("0"))
                newObj.link(newMesh)
                newMesh.update()
            else:
                newLObj = LegoObject('Model', LFile.Name)
                newLObj.ImportFile(lFile=LFile, Invert=Invert)
                newObj = newLObj.Object
            scene.link(newObj)
            newObj.setMatrix(mtxTranf)
            try:
                mat = Material.Get(str(newcolor))
            except:
                mat = None
            if mat != None:
                newObj.setMaterials([mat])
                newObj.colbits=1
            self.Object.makeParent([newObj])

    def AddVertex(self, xyzList, matrix=mtxIdentity):
        vec = Vector(xyzList)
        vec = VecMultMat( vec, matrix.rotationPart())
        vec += matrix.translationPart()
        xyzList = [vec.x, vec.y, vec.z]
        if xyzList in self.lstVertex:
            idx = self.lstVertex.index(xyzList)
        else:
            self.lstVertex.append(xyzList)
            v = NMesh.Vert(vec.x, vec.y, vec.z)
            self.Mesh.verts.append(v)
            idx=len(self.Mesh.verts)-1
        return idx

    def AddTriangle(self, cmd, matrix=mtxIdentity, color=16, CW=False):
        if cmd[0] != '3':       # Check if line contains correct command
            return
        if len(cmd) != 11:      # Check for correct number of parameters
            return
        if int(cmd[1]) != 16:
            color = int(cmd[1])
        f = NMesh.Face()
        if CW :
           list = [2,1,0]
        else:
           list = [0,1,2]
        for x in list:
            idx = self.AddVertex([float(cmd[2+(x*3)]), float(cmd[3+(x*3)]),
                             float(cmd[4+(x*3)])], matrix)
            f.v.append(self.Mesh.verts[idx])
        f.smooth = 1
# Next 3 lines are commented as not completely working way to inport
# multicolos parts
#        idx = self.AddMeshMaterial(color)
#        if idx != None:
#            f.mat = idx
        self.Mesh.faces.append(f)

    def AddQuadrilateral(self, cmd, matrix=mtxIdentity, color=16, CW=False):
        if cmd[0] != '4':       # Check if line contains correct command
            return
        if len(cmd) != 14:      # Check for correct number of parameters
            return
        if int(cmd[1]) != 16:
            color = int(cmd[1])
        f = NMesh.Face()
        if CW :
           list = [3,2,1,0]
        else:
           list = [0,1,2,3]
        for x in list:
            idx = self.AddVertex([float(cmd[2+(x*3)]), float(cmd[3+(x*3)]),
                             float(cmd[4+(x*3)])], matrix)
            f.v.append(self.Mesh.verts[idx])
        f.smooth = 1
# Next 3 lines are commented as not completely working way to inport
# multicolos parts
#        idx = self.AddMeshMaterial(color)
#        if idx != None:
#            f.mat = idx
        self.Mesh.faces.append(f)

    def AddMeshMaterial(self, color):
        if color == 16:
            return None
        if self.MeshMaterials.has_key(color):
            return self.MeshMaterials[color]
        else:
            if len(self.MeshMaterials) == 16:
                return None
            try:
                mat = Blender.Material.Get(str(color))
            except:
                mat = None
            if mat == None:
                return None
            self.Mesh.materials.append(mat)
            idx = len(self.Mesh.materials) - 1
            self.MeshMaterials[color]=idx
            return idx


#=========================================================================



#//////////////////////////////// COLOR TABLE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
#/////
#/////  Values from Official "Ldraw Color Chart"
#/////  Dictionary of  ColorNumber:[Red, Green, Blu, Alpha]
#/////
legocolor={  0: [  51,  51,  51, 255],    #   0  Black LDraw
             1: [   0,  51, 178, 255],    #   1  Blue LDraw
             2: [   0, 127,  51, 255],    #   2  Green LDraw
             3: [   0, 153, 159, 255],    #   3  Dark Cyan LDraw
             4: [ 204,   0,   0, 255],    #   4  Red LDraw
             5: [ 223, 102, 149, 255],    #   5  Magenta LDraw
             6: [ 102,  51,   0, 255],    #   6  Brown LDraw
             7: [ 153, 153, 153, 255],    #   7  Grey LDraw
             8: [ 102, 102,  88, 255],    #   8  Dark Grey LDraw
             9: [   0, 128, 255, 255],    #   9  Light Blue LDraw
            10: [  51, 255, 102, 255],    #  10  Light Green LDraw
            11: [  51, 166, 167, 255],    #  11  Cyan LDraw
            12: [ 255, 201, 196, 255],    #  12  Light Red LDraw
            13: [ 255, 176, 204, 255],    #  13  Pink LDraw
            14: [ 255, 229,   0, 255],    #  14  Yellow LDraw
            15: [ 255, 255, 255, 255],    #  15  White LDraw
            17: [ 102, 240, 153, 255],
            18: [ 204, 142, 104, 255],
            20: [ 224, 204, 240, 255],
            21: [ 196,  40,  27, 255],
            22: [ 153,  51, 153, 255],
            23: [  76,   0, 204, 255],
            24: [ 245, 205,  47, 255],
            25: [  98,  71,  50, 255],
            26: [  27,  42,  52, 255],
            27: [ 229, 255, 168, 255],
            28: [ 250, 163, 157, 255],
            29: [ 161, 196, 139, 255],
            32: [  51,  51,  51, 127],    #  32  Transparent Black LDraw
            33: [   0,  51, 178, 127],    #  33  Transparent Blue LDraw
            34: [   0, 127,  51, 127],    #  34  Transparent Green LDraw
            35: [   0, 153, 159, 127],    #  35  Transparent Dark Cyan LDraw
            36: [ 204,   0,   0, 127],    #  36  Transparent Red LDraw
            37: [ 223, 102, 149, 127],    #  37  Transparent Magenta LDraw
            38: [ 102,  51,   0, 127],    #  38  Transparent Brown LDraw
            39: [ 153, 153, 153, 127],    #  39  Transparent Grey LDraw
            40: [ 102, 102,  88, 127],    #  40  Transparent Dark Grey LDraw
            41: [   0, 128, 255, 127],    #  41  Transparent Light Blue LDraw
            42: [  51, 255, 102, 127],    #  42  Transparent Light Green LDraw
            43: [  51, 166, 167, 127],    #  43  Transparent Cyan LDraw
            44: [ 255, 201, 196, 127],    #  44  Transparent Light Red LDraw
            45: [ 255, 176, 204, 127],    #  45  Transparent Pink LDraw
            46: [ 255, 229,   0, 127],    #  46  Transparent Yellow LDraw
            47: [ 255, 255, 255, 127],    #  47  Transparent White LDraw
            48: [ 132, 182, 141, 255],
            49: [ 248, 241, 132, 255],
            50: [ 236, 232, 222, 255],
           100: [ 238, 196, 182, 255],
           101: [ 218, 134, 121, 255],
           102: [ 110, 153, 201, 255],
           103: [ 199, 193, 183, 255],
           104: [ 107,  50, 123, 255],
           105: [ 226, 155,  63, 255],
           106: [ 218, 133,  64, 255],
           107: [   0, 143, 155, 255],
           108: [ 104,  92,  67, 255],
           110: [  67,  84, 147, 255],
           111: [ 191, 183, 177, 255],
           112: [ 104, 116, 172, 255],
           113: [ 228, 173, 200, 255],
           115: [ 199, 210,  60, 255],
           116: [  85, 165, 175, 255],
           118: [ 183, 215, 213, 255],
           119: [ 164, 189,  70, 255],
           120: [ 217, 228, 167, 255],
           121: [ 231, 172,  88, 255],
           123: [ 211, 111,  76, 255],
           124: [ 146,  57, 120, 255],
           125: [ 234, 184, 145, 255],
           126: [ 165, 165, 203, 255],
           127: [ 220, 188, 129, 255],
           128: [ 174, 122,  89, 255],
           131: [ 156, 163, 168, 255],
           133: [ 213, 115,  61, 255],
           134: [ 216, 221,  86, 255],
           135: [ 116, 134, 156, 255],
           136: [ 135, 124, 144, 255],
           137: [ 224, 152, 100, 255],
           138: [ 149, 138, 115, 255],
           140: [  32,  58,  86, 255],
           141: [  39,  70,  44, 255],
           143: [ 207, 226, 247, 255],
           145: [ 121, 136, 161, 255],
           146: [ 149, 142, 163, 255],
           147: [ 147, 135, 103, 255],
           148: [  87,  88,  87, 255],
           149: [  22,  29,  50, 255],
           150: [ 171, 173, 172, 255],
           151: [ 120, 144, 129, 255],
           153: [ 149, 121, 118, 255],
           154: [ 123,  46,  47, 255],
           157: [ 255, 246, 123, 255],
           158: [ 225, 164, 194, 255],
           168: [ 117, 108,  98, 255],
           176: [ 151, 105,  91, 255],
           178: [ 180, 132,  85, 255],
           179: [ 137, 135, 136, 255],
           180: [ 215, 169,  75, 255],
           190: [ 249, 214,  46, 255],
           191: [ 232, 171,  45, 255],
           192: [ 105,  64,  39, 255],
           193: [ 207,  96,  36, 255],
           194: [ 163, 162, 164, 255],
           195: [  70, 103, 164, 255],
           196: [  35,  71, 139, 255],
           198: [ 142,  66, 133, 255],
           199: [  99,  95,  97, 255],
           200: [ 130, 138,  93, 255],
           208: [ 229, 228, 222, 255],
           209: [ 176, 142,  68, 255],
           210: [ 112, 149, 120, 255],
           211: [ 121, 181, 181, 255],
           212: [ 159, 195, 233, 255],
           213: [ 108, 129, 183, 255],
           216: [ 143,  76,  42, 255],
           217: [ 124,  92,  69, 255],
           218: [ 150, 112, 159, 255],
           219: [ 107,  98, 155, 255],
           220: [ 167, 169, 206, 255],
           221: [ 205,  98, 152, 255],
           222: [ 228, 173, 200, 255],
           223: [ 220, 144, 149, 255],
           224: [ 240, 213, 160, 255],
           225: [ 235, 184, 127, 255],
           226: [ 253, 234, 140, 255],
           232: [ 125, 187, 221, 255],
           268: [  52,  43, 117, 255],
           418: [   0, 191,  89, 255],
           494: [ 204, 204, 204, 255],
           503: [ 230, 227, 218, 255]}
#/////
#/////  Color   16 = Null-colour
#/////        Pieces using null colour code use whatever colour code is
#/////        entered on the command line when rendering with LDraw or,
#/////        if no colour code is specified, they default to a #7 Grey.
#/////  Clear Colours.
#/////        Colors from 32 to 47 are the clear (transparent) version
#/////        of corrispective colors from 0 to 15 (Alpha = 200).
#/////
#//////////////////////////////// COLOR TABLE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*


#//////////////////////////////////// EXITGUI /////////////////////////////////*
#/////
#/////	Exit form the gui
#/////
#//////////////////////////////////// EXITGUI /////////////////////////////////*
def ExitGUI ():
    Exit()

#//////////////////////////////////// EVENTGUI /////////////////////////////////*
#/////
#/////	Process events
#/////
#//////////////////////////////////// EVENTGUI /////////////////////////////////*
def EventGUI (event):
    if (event==EVENT_IMPORT):
        bMaterials = False
        Blender.Window.FileSelector(Import, "LDRAW Import")
    elif (event==EVENT_QUIT):
        ExitGUI()

#//////////////////////////////////// DRAWGUI /////////////////////////////////*
#/////
#/////	Draw the gui in blender
#/////
#//////////////////////////////////// DRAWGUI /////////////////////////////////*
def DrawGUI():
    glClearColor(0.6,0.6,0.6,0)
    glClear(GL_COLOR_BUFFER_BIT)
    glRasterPos2d(100,64)
    Text("Import LDRAW (Lego CAD) file")
    Button ("Import",EVENT_IMPORT,32,10,100,32,"Select File to Import")
    Button ("Cancel",EVENT_QUIT,252,10,100,32,"Do not Import")


#//////////////////////////////////// IMPORT /////////////////////////////////*
#/////
#/////	Load the 3ds file, parse it and import everything in blender
#/////
#//////////////////////////////////// IMPORT /////////////////////////////////*
def Import(FileName):
    print
    print "---------------------START------------------------"
    scene = Blender.Scene.getCurrent ()
    ThisFile = LegoFile(FileName)
    if ThisFile == None:
        print "error while loading Ldraw file make sure path is correct"
        return
    # Create Lego Materials
    for i,LdrawColor in legocolor.iteritems():
        try:
            mat = Blender.Material.Get(str(i))
        except:
            mat = Blender.Material.New(str(i))
        mat.rgbCol = [float(LdrawColor[0])/255,
                      float(LdrawColor[1])/255,
                      float(LdrawColor[2])/255]
        mat.setAlpha( float(LdrawColor[3])/255)
        if LdrawColor[3] < 255:
             mat.mode |= Material.Modes.ZTRANSP

    aObj = LegoObject('Model',ThisFile.Name)

    scene.link (aObj.Object)
    aObj.Object.setLocation(0,0,0)
    aObj.ImportFile(ThisFile)
    aObj.Mesh.update()
    corrMat=Matrix([ Scale, 0    ,     0, 0],
                   [ 0    , 0    ,-Scale, 0],
                   [ 0    , Scale,     0, 0],
                   [ 0    , 0    ,     0, 1])
    aObj.Object.setMatrix(corrMat)
    Blender.Redraw()
    print "-------------------FINISHED-----------------------"
    print

#//////////////////////////////////// REGISTERGUI /////////////////////////////////*
#/////
#/////   Register callbacks for windowing
#/////
#/////   DrawGui  : A function to draw the screen, taking no arguments
#/////   None     : No function to handle keyboard and mouse input events
#/////   EventGUI : A function to handle Draw Button events, taking 'event' argument
#/////
#//////////////////////////////////// REGISTERGUI /////////////////////////////////*
def RegisterGUI ():
    Register (DrawGUI,None,EventGUI)

#////////////////////////////////////  MAIN SCRIPT /////////////////////////////////*
#/////
#/////	Start the script
#/////
#////////////////////////////////////  MAIN SCRIPT /////////////////////////////////*
RegisterGUI()
